@@ -8,6 +8,7 @@ use {
         instruction::InstructionError,
         keyed_account::KeyedAccount,
         pubkey::Pubkey,
+        system_instruction::MAX_PERMITTED_DATA_LENGTH,
     },
     std::{io::prelude::*, mem::size_of},
 };
@@ -27,23 +28,32 @@ pub fn serialize_parameters(
     program_id: &Pubkey,
     keyed_accounts: &[KeyedAccount],
     data: &[u8],
-) -> Result<AlignedMemory, InstructionError> {
+) -> Result<(AlignedMemory, Vec<usize>), InstructionError> {
     if *loader_id == bpf_loader_deprecated::id() {
         serialize_parameters_unaligned(program_id, keyed_accounts, data)
     } else {
         serialize_parameters_aligned(program_id, keyed_accounts, data)
     }
+    .and_then(|buffer| {
+        let account_lengths = keyed_accounts
+            .iter()
+            .map(|keyed_account| keyed_account.data_len())
+            .collect::<Result<Vec<usize>, InstructionError>>()?;
+        Ok((buffer, account_lengths))
+    })
 }
 
 pub fn deserialize_parameters(
     loader_id: &Pubkey,
     keyed_accounts: &[KeyedAccount],
     buffer: &[u8],
+    account_lengths: &[usize],
+    do_support_realloc: bool,
 ) -> Result<(), InstructionError> {
     if *loader_id == bpf_loader_deprecated::id() {
-        deserialize_parameters_unaligned(keyed_accounts, buffer)
+        deserialize_parameters_unaligned(keyed_accounts, buffer, account_lengths)
     } else {
-        deserialize_parameters_aligned(keyed_accounts, buffer)
+        deserialize_parameters_aligned(keyed_accounts, buffer, account_lengths, do_support_realloc)
     }
 }
 
@@ -125,9 +135,14 @@ pub fn serialize_parameters_unaligned(
 pub fn deserialize_parameters_unaligned(
     keyed_accounts: &[KeyedAccount],
     buffer: &[u8],
+    account_lengths: &[usize],
 ) -> Result<(), InstructionError> {
     let mut start = size_of::<u64>(); // number of accounts
-    for (i, keyed_account) in keyed_accounts.iter().enumerate() {
+    for (i, (keyed_account, _pre_len)) in keyed_accounts
+        .iter()
+        .zip(account_lengths.iter())
+        .enumerate()
+    {
         let (is_dup, _) = is_dup(&keyed_accounts[..i], keyed_account);
         start += 1; // is_dup
         if !is_dup {
@@ -246,9 +261,15 @@ pub fn serialize_parameters_aligned(
 pub fn deserialize_parameters_aligned(
     keyed_accounts: &[KeyedAccount],
     buffer: &[u8],
+    account_lengths: &[usize],
+    do_support_realloc: bool,
 ) -> Result<(), InstructionError> {
     let mut start = size_of::<u64>(); // number of accounts
-    for (i, keyed_account) in keyed_accounts.iter().enumerate() {
+    for (i, (keyed_account, pre_len)) in keyed_accounts
+        .iter()
+        .zip(account_lengths.iter())
+        .enumerate()
+    {
         let (is_dup, _) = is_dup(&keyed_accounts[..i], keyed_account);
         start += size_of::<u8>(); // position
         if is_dup {
@@ -264,18 +285,26 @@ pub fn deserialize_parameters_aligned(
             start += size_of::<Pubkey>(); // owner
             account.set_lamports(LittleEndian::read_u64(&buffer[start..]));
             start += size_of::<u64>(); // lamports
-            let pre_len = account.data().len();
             let post_len = LittleEndian::read_u64(&buffer[start..]) as usize;
             start += size_of::<u64>(); // data length
-            let mut data_end = start + pre_len;
-            if post_len != pre_len
-                && (post_len.saturating_sub(pre_len)) <= MAX_PERMITTED_DATA_INCREASE
-            {
-                data_end = start + post_len;
-            }
-
+            let data_end = if do_support_realloc {
+                if post_len.saturating_sub(*pre_len) > MAX_PERMITTED_DATA_INCREASE
+                    || post_len > MAX_PERMITTED_DATA_LENGTH as usize
+                {
+                    return Err(InstructionError::InvalidRealloc);
+                }
+                start + post_len
+            } else {
+                let mut data_end = start + *pre_len;
+                if post_len != *pre_len
+                    && (post_len.saturating_sub(*pre_len)) <= MAX_PERMITTED_DATA_INCREASE
+                {
+                    data_end = start + post_len;
+                }
+                data_end
+            };
             account.set_data_from_slice(&buffer[start..data_end]);
-            start += pre_len + MAX_PERMITTED_DATA_INCREASE; // data
+            start += *pre_len + MAX_PERMITTED_DATA_INCREASE; // data
             start += (start as *const u8).align_offset(BPF_ALIGN_OF_U128);
             start += size_of::<u64>(); // rent_epoch
         }
@@ -287,6 +316,7 @@ pub fn deserialize_parameters_aligned(
 mod tests {
     use {
         super::*,
+        solana_program_runtime::invoke_context::{prepare_mock_invoke_context, InvokeContext},
         solana_sdk::{
             account::{Account, AccountSharedData},
             account_info::AccountInfo,
@@ -305,97 +335,137 @@ mod tests {
         let program_id = solana_sdk::pubkey::new_rand();
         let dup_key = solana_sdk::pubkey::new_rand();
         let dup_key2 = solana_sdk::pubkey::new_rand();
-        let keys = vec![
-            dup_key,
-            dup_key,
-            solana_sdk::pubkey::new_rand(),
-            solana_sdk::pubkey::new_rand(),
-            dup_key2,
-            dup_key2,
-            solana_sdk::pubkey::new_rand(),
-            solana_sdk::pubkey::new_rand(),
+        let keyed_accounts = [
+            (
+                false,
+                false,
+                program_id,
+                Rc::new(RefCell::new(AccountSharedData::from(Account {
+                    lamports: 0,
+                    data: vec![],
+                    owner: bpf_loader::id(),
+                    executable: true,
+                    rent_epoch: 0,
+                }))),
+            ),
+            (
+                false,
+                false,
+                dup_key,
+                Rc::new(RefCell::new(AccountSharedData::from(Account {
+                    lamports: 1,
+                    data: vec![1u8, 2, 3, 4, 5],
+                    owner: bpf_loader::id(),
+                    executable: false,
+                    rent_epoch: 100,
+                }))),
+            ),
+            (
+                false,
+                false,
+                dup_key,
+                Rc::new(RefCell::new(AccountSharedData::from(Account {
+                    lamports: 1,
+                    data: vec![1u8, 2, 3, 4, 5],
+                    owner: bpf_loader::id(),
+                    executable: false,
+                    rent_epoch: 100,
+                }))),
+            ),
+            (
+                false,
+                false,
+                solana_sdk::pubkey::new_rand(),
+                Rc::new(RefCell::new(AccountSharedData::from(Account {
+                    lamports: 2,
+                    data: vec![11u8, 12, 13, 14, 15, 16, 17, 18, 19],
+                    owner: bpf_loader::id(),
+                    executable: true,
+                    rent_epoch: 200,
+                }))),
+            ),
+            (
+                false,
+                false,
+                solana_sdk::pubkey::new_rand(),
+                Rc::new(RefCell::new(AccountSharedData::from(Account {
+                    lamports: 3,
+                    data: vec![],
+                    owner: bpf_loader::id(),
+                    executable: false,
+                    rent_epoch: 3100,
+                }))),
+            ),
+            (
+                false,
+                true,
+                dup_key2,
+                Rc::new(RefCell::new(AccountSharedData::from(Account {
+                    lamports: 4,
+                    data: vec![1u8, 2, 3, 4, 5],
+                    owner: bpf_loader::id(),
+                    executable: false,
+                    rent_epoch: 100,
+                }))),
+            ),
+            (
+                false,
+                true,
+                dup_key2,
+                Rc::new(RefCell::new(AccountSharedData::from(Account {
+                    lamports: 4,
+                    data: vec![1u8, 2, 3, 4, 5],
+                    owner: bpf_loader::id(),
+                    executable: false,
+                    rent_epoch: 100,
+                }))),
+            ),
+            (
+                false,
+                true,
+                solana_sdk::pubkey::new_rand(),
+                Rc::new(RefCell::new(AccountSharedData::from(Account {
+                    lamports: 5,
+                    data: vec![11u8, 12, 13, 14, 15, 16, 17, 18, 19],
+                    owner: bpf_loader::id(),
+                    executable: true,
+                    rent_epoch: 200,
+                }))),
+            ),
+            (
+                false,
+                true,
+                solana_sdk::pubkey::new_rand(),
+                Rc::new(RefCell::new(AccountSharedData::from(Account {
+                    lamports: 6,
+                    data: vec![],
+                    owner: bpf_loader::id(),
+                    executable: false,
+                    rent_epoch: 3100,
+                }))),
+            ),
         ];
-        let accounts = [
-            RefCell::new(AccountSharedData::from(Account {
-                lamports: 1,
-                data: vec![1u8, 2, 3, 4, 5],
-                owner: bpf_loader::id(),
-                executable: false,
-                rent_epoch: 100,
-            })),
-            // dup
-            RefCell::new(AccountSharedData::from(Account {
-                lamports: 1,
-                data: vec![1u8, 2, 3, 4, 5],
-                owner: bpf_loader::id(),
-                executable: false,
-                rent_epoch: 100,
-            })),
-            RefCell::new(AccountSharedData::from(Account {
-                lamports: 2,
-                data: vec![11u8, 12, 13, 14, 15, 16, 17, 18, 19],
-                owner: bpf_loader::id(),
-                executable: true,
-                rent_epoch: 200,
-            })),
-            RefCell::new(AccountSharedData::from(Account {
-                lamports: 3,
-                data: vec![],
-                owner: bpf_loader::id(),
-                executable: false,
-                rent_epoch: 3100,
-            })),
-            RefCell::new(AccountSharedData::from(Account {
-                lamports: 4,
-                data: vec![1u8, 2, 3, 4, 5],
-                owner: bpf_loader::id(),
-                executable: false,
-                rent_epoch: 100,
-            })),
-            // dup
-            RefCell::new(AccountSharedData::from(Account {
-                lamports: 4,
-                data: vec![1u8, 2, 3, 4, 5],
-                owner: bpf_loader::id(),
-                executable: false,
-                rent_epoch: 100,
-            })),
-            RefCell::new(AccountSharedData::from(Account {
-                lamports: 5,
-                data: vec![11u8, 12, 13, 14, 15, 16, 17, 18, 19],
-                owner: bpf_loader::id(),
-                executable: true,
-                rent_epoch: 200,
-            })),
-            RefCell::new(AccountSharedData::from(Account {
-                lamports: 6,
-                data: vec![],
-                owner: bpf_loader::id(),
-                executable: false,
-                rent_epoch: 3100,
-            })),
-        ];
-
-        let keyed_accounts: Vec<_> = keys
-            .iter()
-            .zip(&accounts)
-            .enumerate()
-            .map(|(i, (key, account))| {
-                if i <= accounts.len() / 2 {
-                    KeyedAccount::new_readonly(key, false, account)
-                } else {
-                    KeyedAccount::new(key, false, account)
-                }
-            })
-            .collect();
         let instruction_data = vec![1u8, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11];
+        let program_indices = [0];
+        let preparation =
+            prepare_mock_invoke_context(&program_indices, &instruction_data, &keyed_accounts);
+        let mut invoke_context = InvokeContext::new_mock(&preparation.accounts, &[]);
+        invoke_context
+            .push(
+                &preparation.message,
+                &preparation.message.instructions[0],
+                &program_indices,
+                &preparation.account_indices,
+            )
+            .unwrap();
 
         // check serialize_parameters_aligned
 
-        let mut serialized = serialize_parameters(
+        let ser_keyed_accounts = invoke_context.get_keyed_accounts().unwrap();
+        let (mut serialized, account_lengths) = serialize_parameters(
             &bpf_loader::id(),
             &program_id,
-            &keyed_accounts,
+            &ser_keyed_accounts[1..],
             &instruction_data,
         )
         .unwrap();
@@ -409,8 +479,8 @@ mod tests {
             (&de_instruction_data[0] as *const u8).align_offset(BPF_ALIGN_OF_U128),
             0
         );
-        for ((account, account_info), key) in accounts.iter().zip(de_accounts).zip(keys.clone()) {
-            assert_eq!(key, *account_info.key);
+        for ((_, _, key, account), account_info) in keyed_accounts.iter().skip(1).zip(de_accounts) {
+            assert_eq!(key, account_info.key);
             let account = account.borrow();
             assert_eq!(account.lamports(), account_info.lamports());
             assert_eq!(account.data(), &account_info.data.borrow()[..]);
@@ -432,25 +502,18 @@ mod tests {
             );
         }
 
-        let de_accounts = accounts.clone();
-        let de_keyed_accounts: Vec<_> = keys
-            .iter()
-            .zip(&de_accounts)
-            .enumerate()
-            .map(|(i, (key, account))| {
-                if i <= accounts.len() / 2 {
-                    KeyedAccount::new_readonly(key, false, account)
-                } else {
-                    KeyedAccount::new(key, false, account)
-                }
-            })
-            .collect();
-        deserialize_parameters(&bpf_loader::id(), &de_keyed_accounts, serialized.as_slice())
-            .unwrap();
-        for ((account, de_keyed_account), key) in
-            accounts.iter().zip(de_keyed_accounts).zip(keys.clone())
+        let de_keyed_accounts = invoke_context.get_keyed_accounts().unwrap();
+        deserialize_parameters(
+            &bpf_loader::id(),
+            &de_keyed_accounts[1..],
+            serialized.as_slice(),
+            &account_lengths,
+            true,
+        )
+        .unwrap();
+        for ((_, _, key, account), de_keyed_account) in keyed_accounts.iter().zip(de_keyed_accounts)
         {
-            assert_eq!(key, *de_keyed_account.unsigned_key());
+            assert_eq!(key, de_keyed_account.unsigned_key());
             let account = account.borrow();
             assert_eq!(account.executable(), de_keyed_account.executable().unwrap());
             assert_eq!(account.rent_epoch(), de_keyed_account.rent_epoch().unwrap());
@@ -458,10 +521,11 @@ mod tests {
 
         // check serialize_parameters_unaligned
 
-        let mut serialized = serialize_parameters(
+        let ser_keyed_accounts = invoke_context.get_keyed_accounts().unwrap();
+        let (mut serialized, account_lengths) = serialize_parameters(
             &bpf_loader_deprecated::id(),
             &program_id,
-            &keyed_accounts,
+            &ser_keyed_accounts[1..],
             &instruction_data,
         )
         .unwrap();
@@ -470,8 +534,8 @@ mod tests {
             unsafe { deserialize_unaligned(&mut serialized.as_slice_mut()[0] as *mut u8) };
         assert_eq!(&program_id, de_program_id);
         assert_eq!(instruction_data, de_instruction_data);
-        for ((account, account_info), key) in accounts.iter().zip(de_accounts).zip(keys.clone()) {
-            assert_eq!(key, *account_info.key);
+        for ((_, _, key, account), account_info) in keyed_accounts.iter().skip(1).zip(de_accounts) {
+            assert_eq!(key, account_info.key);
             let account = account.borrow();
             assert_eq!(account.lamports(), account_info.lamports());
             assert_eq!(account.data(), &account_info.data.borrow()[..]);
@@ -480,29 +544,18 @@ mod tests {
             assert_eq!(account.rent_epoch(), account_info.rent_epoch);
         }
 
-        let de_accounts = accounts.clone();
-        let de_keyed_accounts: Vec<_> = keys
-            .iter()
-            .zip(&de_accounts)
-            .enumerate()
-            .map(|(i, (key, account))| {
-                if i < accounts.len() / 2 {
-                    KeyedAccount::new_readonly(key, false, account)
-                } else {
-                    KeyedAccount::new(key, false, account)
-                }
-            })
-            .collect();
+        let de_keyed_accounts = invoke_context.get_keyed_accounts().unwrap();
         deserialize_parameters(
             &bpf_loader_deprecated::id(),
-            &de_keyed_accounts,
+            &de_keyed_accounts[1..],
             serialized.as_slice(),
+            &account_lengths,
+            true,
         )
         .unwrap();
-        for ((account, de_keyed_account), key) in
-            accounts.iter().zip(de_keyed_accounts).zip(keys.clone())
+        for ((_, _, key, account), de_keyed_account) in keyed_accounts.iter().zip(de_keyed_accounts)
         {
-            assert_eq!(key, *de_keyed_account.unsigned_key());
+            assert_eq!(key, de_keyed_account.unsigned_key());
             let account = account.borrow();
             assert_eq!(account.lamports(), de_keyed_account.lamports().unwrap());
             assert_eq!(
